/***************************************************************************
 *                                                                         *
 *    LIBDSK: General floppy and diskimage access library                  *
 *    Copyright (C) 2001  John Elliott <jce@seasip.demon.co.uk>            *
 *                                                                         *
 *    This library is free software; you can redistribute it and/or        *
 *    modify it under the terms of the GNU Library General Public          *
 *    License as published by the Free Software Foundation; either         *
 *    version 2 of the License, or (at your option) any later version.     *
 *                                                                         *
 *    This library is distributed in the hope that it will be useful,      *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *    Library General Public License for more details.                     *
 *                                                                         *
 *    You should have received a copy of the GNU Library General Public    *
 *    License along with this library; if not, write to the Free           *
 *    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,      *
 *    MA 02111-1307, USA                                                   *
 *                                                                         *
 ***************************************************************************/

/* Access functions for compressed CPCEMU discs */

#include "compi.h"
#include "compsq.h"

/* Unsqueeze engine functions */
static void init_cr(SQ_COMPRESS_DATA *self);
static void init_dehuff(SQ_COMPRESS_DATA *self);
static int getcr(SQ_COMPRESS_DATA *self, FILE *ib);
static int getuhuff(SQ_COMPRESS_DATA *self, FILE *ib);
static dsk_err_t getx16(SQ_COMPRESS_DATA *self, FILE *iob, unsigned short *w);    
static dsk_err_t getw16(SQ_COMPRESS_DATA *self, FILE *iob, signed short *w);    
static dsk_err_t unsqueeze(SQ_COMPRESS_DATA *self, FILE *fpin, FILE *fpout);
static dsk_err_t usq_bytes(SQ_COMPRESS_DATA *self, FILE *fpin, FILE *fpout);

/* Squeeze engine calls */
#define NOHIST	0 	/*don't consider previous input*/
#define SENTCHAR 1 	/*lastchar set, no lookahead yet */
#define SENDNEWC 2 	/*newchar set, previous sequence done */
#define SENDCNT 3 	/*newchar set, DLE sent, send count next */

/* *** Stuff for second translation module *** */

#define NOCHILD -1	/* indicates end of path through tree */
#define ERROR   -1

#define MAXCOUNT 0xFFFF	/* biggest unsigned integer */

static dsk_err_t squeeze(SQ_COMPRESS_DATA *self, FILE *fpin, FILE *fpout);
static void init_ncr(SQ_COMPRESS_DATA *self);
static void init_huff(SQ_COMPRESS_DATA *self, FILE *fpin);
static dsk_err_t wrt_head(SQ_COMPRESS_DATA *self, FILE *ob);
static int gethuff(SQ_COMPRESS_DATA *self, FILE *fpin);

static dsk_err_t putwe(SQ_COMPRESS_DATA *self, int w,  FILE *iob);
static void init_tree(SQ_COMPRESS_DATA *self);
static void init_enc(SQ_COMPRESS_DATA *self);
static void bld_tree(SQ_COMPRESS_DATA *self, int list[], int len);
static void adjust(SQ_COMPRESS_DATA *self, int list[], int top, int bottom);
static void heap(SQ_COMPRESS_DATA *self, int list[], int length);
static void scale(SQ_COMPRESS_DATA *self, unsigned int ceil);
static int getcnr(SQ_COMPRESS_DATA *self, FILE *fpin);
static int buildenc(SQ_COMPRESS_DATA *self, int level, int root);
static char cmptrees(SQ_COMPRESS_DATA *self, int a, int b);


/* This struct contains function pointers to the driver's functions, and the
 * size of its DSK_DRIVER subclass */

COMPRESS_CLASS cc_sq = 
{
	sizeof(SQ_COMPRESS_DATA),
	"sq",
	"Squeeze (Huffman coding)",
	sq_open,	/* open */
	sq_creat,	/* create new */
	sq_commit,	/* commit */
	sq_abort	/* abort */
};



dsk_err_t sq_open(COMPRESS_DATA *self)
{
	SQ_COMPRESS_DATA *sq_self;
	FILE *fp, *fpout = NULL;
	dsk_err_t err;
	unsigned short magic;

	
	/* Sanity check: Is this meant for our driver? */
	if (self->cd_class != &cc_sq) return DSK_ERR_BADPTR;
	sq_self = (SQ_COMPRESS_DATA *)self;
	sq_self->sq_truename = NULL;

	/* Open the file to decompress */
	err = comp_fopen(self, &fp);
	if (err) return DSK_ERR_NOTME;

	/* Check for SQ signature */
	err = getx16(sq_self, fp, &magic);
	if (err) return err;

	if (magic != RECOGNIZE)
	{
		fclose(fp);
		return DSK_ERR_NOTME;
	}
	/* OK. This is a Squeezed file. Decompress it. */
	rewind(fp);
	err = comp_mktemp(self, &fpout);
	if (!err) err = unsqueeze(sq_self, fp, fpout);
	fclose(fp);
	if (fpout) fclose(fpout);
	if (err && sq_self->sq_truename) free(sq_self->sq_truename);
	return err;	
}


dsk_err_t sq_commit(COMPRESS_DATA *self)
{
	SQ_COMPRESS_DATA *sq_self;
	dsk_err_t err = DSK_ERR_OK;
	FILE *fpin = NULL, *fpout = NULL;
	
	if (self->cd_class != &cc_sq) return DSK_ERR_BADPTR;
	sq_self = (SQ_COMPRESS_DATA *)self;

        if (self->cd_cfilename && self->cd_ufilename)
        {
		fpin  = fopen(self->cd_ufilename, "rb");
		fpout = fopen(self->cd_cfilename, "wb");
		if (!fpin || !fpout) err = DSK_ERR_SYSERR;
		else err = squeeze(sq_self, fpin, fpout);
	}
	if (fpin) fclose(fpin);
	if (fpout) fclose(fpout);
	if (sq_self->sq_truename) free(sq_self->sq_truename);
	sq_self->sq_truename = NULL;

	return err;
}


/* abort: Free truename */
dsk_err_t sq_abort(COMPRESS_DATA *self)
{
	SQ_COMPRESS_DATA *sq_self;
	if (self->cd_class != &cc_sq) return DSK_ERR_BADPTR;
	sq_self = (SQ_COMPRESS_DATA *)self;

	if (sq_self->sq_truename) free(sq_self->sq_truename);
	sq_self->sq_truename = NULL;
	return DSK_ERR_OK;
}

/* Create: Set up truename */
dsk_err_t sq_creat(COMPRESS_DATA *self)
{
	char *ss;
	SQ_COMPRESS_DATA *sq_self;
	if (self->cd_class != &cc_sq) return DSK_ERR_BADPTR;
	sq_self = (SQ_COMPRESS_DATA *)self;

	if (sq_self->sq_truename) free(sq_self->sq_truename);
	sq_self->sq_truename = NULL;

/* Try to guess the true name of the file from the compressed name. This 
 * can only be done for certain well-known name manglings. */
        sq_self->sq_truename = malloc(1 + strlen(self->cd_cfilename));
        if (!sq_self->sq_truename) return DSK_ERR_NOMEM;

	/* UNIX SQ squeezes files by appending ".SQ" */
        strcpy(sq_self->sq_truename, self->cd_cfilename);
	ss = strstr(sq_self->sq_truename, ".SQ");
	if (ss) *ss = 0;

	/* Convert .DQK back to .DSK */
	ss = strstr(sq_self->sq_truename, ".DQK");
	if (ss) memcpy(ss, ".DSK", 4);

	/* Convert .DQK back to .DSK */
	ss = strstr(sq_self->sq_truename, ".dqk");
	if (ss) memcpy(ss, ".dsk", 4);

	return DSK_ERR_OK;
}


dsk_err_t unsqueeze(SQ_COMPRESS_DATA *self, FILE *inbuff, FILE *outbuff)
{
	int i, c;

	unsigned short filecrc;	/* checksum */
	unsigned short numnodes;	/* size of decoding tree */
	int err;
	long lpos1, lpos2;
	unsigned short magic;

	/* Initialization */
	self->crc = 0;
	init_cr(self);
	init_dehuff(self);

	/* Process header */
	err = getx16(self, inbuff, &magic);
	if (err) return err;
	if(magic != RECOGNIZE) return DSK_ERR_NOTME;

	err = getx16(self, inbuff, &filecrc);
	if (err) return err;
	/* Get original file name */

	lpos1 = ftell(inbuff);
	do
	{
		c = fgetc(inbuff);
		if (c == EOF) return DSK_ERR_COMPRESS;
	}
	while (c != 0);

	lpos2 = ftell(inbuff);
	self->sq_truename = malloc(lpos2 - lpos1);
	if (!self->sq_truename) return DSK_ERR_COMPRESS;
	fseek(inbuff, lpos1, SEEK_SET);
	fread(self->sq_truename, 1, lpos2 - lpos1, inbuff);

	err = getx16(self, inbuff, &numnodes);
	if (err && self->sq_truename) free(self->sq_truename);
	if (err) return err;
	
	if(numnodes >= NUMVALS) 
	{
		/* Invalid decode tree size */
		return DSK_ERR_COMPRESS;
	}

	/* Initialize for possible empty tree (SPEOF only) */
	self->dnode[0].children[0] = -(SPEOF + 1);
	self->dnode[0].children[1] = -(SPEOF + 1);

	/* Get decoding tree from file */
	for(i = 0; i < numnodes; ++i) 
	{
		err = getw16(self, inbuff, &self->dnode[i].children[0]);
		if (err) return err;
		err = getw16(self, inbuff, &self->dnode[i].children[1]);
		if (err) return err;
	}

	/* Do decompression */
	err = usq_bytes(self, inbuff, outbuff);
	if (!err)
	{	/* Checksum error */
		if((filecrc && 0xFFFF) != (self->crc && 0xFFFF))
			return DSK_ERR_COMPRESS;
	}
	return err;
}

/* Decompress until EOF */
static dsk_err_t usq_bytes(SQ_COMPRESS_DATA *self, FILE *fpin, FILE *fpout)
{
	int c;

	while((c = getcr(self, fpin)) != EOF) 
	{
		self->crc += c;
		fputc(c, fpout);
	}
	return DSK_ERR_OK;
}


/* get 16-bit unsigned word from file */
dsk_err_t getx16(SQ_COMPRESS_DATA *self, FILE *iob, unsigned short *v)
{
	int temp;
	unsigned short rv;

	temp = fgetc(iob);		/* get low order byte */
	if (temp == EOF) return DSK_ERR_NOTME;
	rv = temp;
	temp = fgetc(iob);		/* get high order byte */
	if (temp == EOF) return DSK_ERR_NOTME;
	rv |= (temp << 8);
	*v = rv;
	return DSK_ERR_OK;
}

/* get 16-bit signed word from file */
dsk_err_t getw16(SQ_COMPRESS_DATA *self, FILE *iob, signed short *v)
{
	unsigned short u;
	dsk_err_t err = getx16(self, iob, &u);
	if (err) return err;

	if (u & 0x8000)
	{
		int i = (u & 0x7FFF) - 0x8000;
		*v = i;
	}
	else *v = u;
	return DSK_ERR_OK;

}



void init_cr(SQ_COMPRESS_DATA *self)
{
	self->repct = 0;
}

void init_dehuff(SQ_COMPRESS_DATA *self)
{
	self->bpos = 99;	/* force initial read */
}

/* Get bytes with decoding - this decodes repetition,
 * calls getuhuff to decode file stream into byte
 * level code with only repetition encoding.
 *
 * The code is simple passing through of bytes except
 * that DLE is encoded as DLE-zero and other values
 * repeated more than twice are encoded as value-DLE-count.
 */

int getcr(SQ_COMPRESS_DATA *self, FILE *ib)
{
	int c;

	if(self->repct > 0) {
		/* Expanding a repeated char */
		--self->repct;
		return self->value;
	} else {
		/* Nothing unusual */
		if((c = getuhuff(self, ib)) != DLE) {
			/* It's not the special delimiter */
			self->value = c;
			if(self->value == EOF)
				self->repct = LARGE;
			return self->value;
		} else {
			/* Special token */
			if((self->repct = getuhuff(self, ib)) == 0)
				/* DLE, zero represents DLE */
				return DLE;
			else {
				/* Begin expanding repetition */
				self->repct -= 2;	/* 2nd time */
				return self->value;
			}
		}
	}
	return EOF;
}

/* ejecteject */

/* Decode file stream into a byte level code with only
 * repetition encoding remaining.
 */

int getuhuff(SQ_COMPRESS_DATA *self, FILE *ib)
{
	int i;

	/* Follow bit stream in tree to a leaf*/
	i = 0;	/* Start at root of tree */
	do {
		if(++self->bpos > 7) {
			if((self->curin = fgetc(ib)) == EOF)
				return EOF;
			self->bpos = 0;
			/* move a level deeper in tree */
			i = self->dnode[i].children[1 & self->curin];
		} else
			i = self->dnode[i].children[1 & (self->curin >>= 1)];
	} while(i >= 0);

	/* Decode fake node index to original data value */
	i = -(i + 1);
	/* Decode special endfile token to normal EOF */
	i = (i == SPEOF) ? EOF : i;
	return i;
}


static dsk_err_t squeeze(SQ_COMPRESS_DATA *self, FILE *fpin, FILE *iob)
{
	int c;
	dsk_err_t err;

	self->crc = 0;
	/* First pass: Analyze */
	init_ncr(self);
	init_huff(self, fpin);
	rewind(fpin);
	/* Write output header */	
	err = wrt_head(self, iob); 
	if (err) return err;

	/* Second pass: encode the file */

	init_ncr(self);	/* For second pass */

	/* Translate the input file into the output file */
	while((c = gethuff(self, fpin)) != EOF)
	{
		err = (fputc(c, iob) == EOF);
		if (err) return DSK_ERR_SYSERR; 
	}
	return DSK_ERR_OK;
}


static int getc_crc(SQ_COMPRESS_DATA *self, FILE *fpin)
{
	int ch;
	
	ch = fgetc(fpin);
	if (ch == EOF) return ch;
	self->crc += ch;
	return ch;
}



void init_ncr(SQ_COMPRESS_DATA *self)	/*initialize getcnr() */
{
	self->state = NOHIST;
}

int getcnr(SQ_COMPRESS_DATA *self, FILE *fpin)
{
	switch(self->state) 
	{
	case NOHIST:
		/* No relevant history */
		self->state = SENTCHAR;
		return self->lastchar = getc_crc(self, fpin);   
	case SENTCHAR:
		/* Lastchar is set, need lookahead */
		switch(self->lastchar) 
		{
		case DLE:
			self->state = NOHIST;
			return 0;	/* indicates DLE was the data */
		case EOF:
			return EOF;
		default:
			for(self->likect = 1; (self->newchar = getc_crc(self, fpin)) == self->lastchar && self->likect < 255; ++self->likect)
				;
			switch(self->likect) {
			case 1:
				return self->lastchar = self->newchar;
			case 2:
				/* just pass through */
				self->state = SENDNEWC;
				return self->lastchar;
			default:
				self->state = SENDCNT;
				return DLE;
			}
		}
	case SENDNEWC:
		/* Previous sequence complete, newchar set */
		self->state = SENTCHAR;
		return self->lastchar = self->newchar;
	case SENDCNT:
		/* Sent DLE for repeat sequence, send count */
		self->state = SENDNEWC;
		return self->likect;
	default:
		fputs("LIBDSK internal error: SQ Bug - bad state\n", stderr);
		exit(1);
	}
}


/******** Second translation - bytes to variable length bit strings *********/


/* This translation uses the Huffman algorithm to develop a
 * binary tree representing the decoding information for
 * a variable length bit string code for each input value.
 * Each string's length is in inverse proportion to its
 * frequency of appearance in the incoming data stream.
 * The encoding table is derived from the decoding table.
 *
 * The range of valid values into the Huffman algorithm are
 * the values of a byte stored in an integer plus the special
 * endfile value chosen to be an adjacent value. Overall, 0-SPEOF.
 *
 * The "node" array of structures contains the nodes of the
 * binary tree. The first NUMVALS nodes are the leaves of the
 * tree and represent the values of the data bytes being
 * encoded and the special endfile, SPEOF.
 * The remaining nodes become the internal nodes of the tree.
 *
 * In the original design it was believed that
 * a Huffman code would fit in the same number of
 * bits that will hold the sum of all the counts.
 * That was disproven by a user's file and was a rare but
 * infamous bug. This version attempts to choose among equally
 * weighted subtrees according to their maximum depths to avoid
 * unnecessarily long codes. In case that is not sufficient
 * to guarantee codes <= 16 bits long, we initially scale
 * the counts so the total fits in an unsigned integer, but
 * if codes longer than 16 bits are generated the counts are
 * rescaled to a lower ceiling and code generation is retried.
 */

/* Initialize the Huffman translation. This requires reading
 * the input file through any preceding translation functions
 * to get the frequency distribution of the various values.
 */

void init_huff(SQ_COMPRESS_DATA *self, FILE *fpin)          
{
	int c, i;
	int btlist[NUMVALS];	/* list of intermediate binary trees */
	int listlen;		/* length of btlist */
	unsigned int *wp;	/* simplifies weight counting */
	unsigned int ceiling;	/* limit for scaling */

	/* Initialize tree nodes to no weight, no children */
	init_tree(self);

	/* Build frequency info in tree */
	do {
		c = getcnr(self, fpin);        
		if(c == EOF)
			c = SPEOF;
		if(*(wp = &self->node[c].weight) !=  MAXCOUNT)
			++(*wp);
	} while(c != SPEOF);
	ceiling = MAXCOUNT;

	do {	/* Keep trying to scale and encode */
		scale(self, ceiling);
		ceiling /= 2;	/* in case we rescale */

		/* Build list of single node binary trees having
		 * leaves for the input values with non-zero counts
		 */
		for(i = listlen = 0; i < NUMVALS; ++i)
			if(self->node[i].weight != 0) {
				self->node[i].tdepth = 0;
				btlist[listlen++] = i;
			}

		/* Arrange list of trees into a heap with the entry
		 * indexing the node with the least weight at the top.
		 */
		heap(self, btlist, listlen);

		/* Convert the list of trees to a single decoding tree */
		bld_tree(self, btlist, listlen);

		/* Initialize the encoding table */
		init_enc(self);

		/* Try to build encoding table.
		 * Fail if any code is > 16 bits long.
		 */
	} while(buildenc(self, 0, self->dctreehd) == ERROR);

	/* Initialize encoding variables */
	self->cbitsrem = 0;	/*force initial read */
	self->curin = 0;	/*anything but endfile*/
}
/*  */
/* The count of number of occurrances of each input value
 * have already been prevented from exceeding MAXCOUNT.
 * Now we must scale them so that their sum doesn't exceed
 * ceiling and yet no non-zero count can become zero.
 * This scaling prevents errors in the weights of the
 * interior nodes of the Huffman tree and also ensures that
 * the codes will fit in an unsigned integer. Rescaling is
 * used if necessary to limit the code length.
 */

void scale(SQ_COMPRESS_DATA *self, unsigned int ceil)
{
	int ovflw, divisor, i;
	unsigned int w, sum;
	char increased;		/* flag */

	do {
		for(i = sum = ovflw = 0; i < NUMVALS; ++i) {
			if(self->node[i].weight > (ceil - sum))
				++ovflw;
			sum += self->node[i].weight;
		}

		divisor = ovflw + 1;

		/* Ensure no non-zero values are lost */
		increased = 0;
		for(i = 0; i < NUMVALS; ++i) {
			w = self->node[i].weight;
			if (w < divisor && w > 0) {
				/* Don't fail to provide a code if it's used at all */
				self->node[i].weight = divisor;
				increased = 1;
			}
		}
	} while(increased);

	/* Scaling factor choosen, now scale */
	if(divisor > 1)
		for(i = 0; i < NUMVALS; ++i)
			self->node[i].weight /= divisor;
}
/*  */
/* heap() and adjust() maintain a list of binary trees as a
 * heap with the top indexing the binary tree on the list
 * which has the least weight or, in case of equal weights,
 * least depth in its longest path. The depth part is not
 * strictly necessary, but tends to avoid long codes which
 * might provoke rescaling.
 */

void heap(SQ_COMPRESS_DATA *self, int list[], int length)
{
	int i;

	for(i = (length - 2) / 2; i >= 0; --i)
		adjust(self, list, i, length - 1);
}

/* Make a heap from a heap with a new top */

void adjust(SQ_COMPRESS_DATA *self, int list[], int top, int bottom)
{
	int k, temp;

	k = 2 * top + 1;	/* left child of top */
	temp = list[top];	/* remember root node of top tree */
	if( k <= bottom) {
		if( k < bottom && cmptrees(self, list[k], list[k + 1]))
			++k;

		/* k indexes "smaller" child (in heap of trees) of top */
		/* now make top index "smaller" of old top and smallest child */
		if(cmptrees(self, temp, list[k])) {
			list[top] = list[k];
			list[k] = temp;
			/* Make the changed list a heap */
			adjust(self, list, k, bottom); /*recursive*/
		}
	}
}

/* Compare two trees, if a > b return true, else return false
 * note comparison rules in previous comments.
 */

char cmptrees(SQ_COMPRESS_DATA *self, int a, int b)
{
	if(self->node[a].weight > self->node[b].weight)
		return 1;
	if(self->node[a].weight == self->node[b].weight)
		if(self->node[a].tdepth > self->node[b].tdepth)
			return 1;
	return 0;
}

static char maxchar(char a, char b)
{
	return a > b ? a : b;
}
/*  */
/* HUFFMAN ALGORITHM: develops the single element trees
 * into a single binary tree by forming subtrees rooted in
 * interior nodes having weights equal to the sum of weights of all
 * their descendents and having depth counts indicating the
 * depth of their longest paths.
 *
 * When all trees have been formed into a single tree satisfying
 * the heap property (on weight, with depth as a tie breaker)
 * then the binary code assigned to a leaf (value to be encoded)
 * is then the series of left (0) and right (1)
 * paths leading from the root to the leaf.
 * Note that trees are removed from the heaped list by
 * moving the last element over the top element and
 * reheaping the shorter list.
 */

void bld_tree(SQ_COMPRESS_DATA *self, int list[], int len)
{
	int freenode;		/* next free node in tree */
	int lch, rch;		/* temporaries for left, right children */
	struct nd *frnp;	/* free node pointer */

	/* Initialize index to next available (non-leaf) node.
	 * Lower numbered nodes correspond to leaves (data values).
	 */
	freenode = NUMVALS;

	while(len > 1) {
		/* Take from list two btrees with least weight
		 * and build an interior node pointing to them.
		 * This forms a new tree.
		 */
		lch = list[0];	/* This one will be left child */

		/* delete top (least) tree from the list of trees */
		list[0] = list[--len];
		adjust(self, list, 0, len - 1);

		/* Take new top (least) tree. Reuse list slot later */
		rch = list[0];	/* This one will be right child */

		/* Form new tree from the two least trees using
		 * a free node as root. Put the new tree in the list.
		 */
		frnp = &self->node[freenode];	/* address of next free node */
		list[0] = freenode++;	/* put at top for now */
		frnp->lchild = lch;
		frnp->rchild = rch;
		frnp->weight = self->node[lch].weight + self->node[rch].weight;
		frnp->tdepth = 1 + maxchar(self->node[lch].tdepth, self->node[rch].tdepth);
		/* reheap list  to get least tree at top*/
		adjust(self, list, 0, len - 1);
	}
	self->dctreehd = list[0];	/*head of final tree */
}
/* Initialize all nodes to single element binary trees
 * with zero weight and depth.
 */

void init_tree(SQ_COMPRESS_DATA *self)
{
	int i;

	for(i = 0; i < NUMNODES; ++i) {
		self->node[i].weight = 0;
		self->node[i].tdepth = 0;
		self->node[i].lchild = NOCHILD;
		self->node[i].rchild = NOCHILD;
	}
}

void init_enc(SQ_COMPRESS_DATA *self)
{
	int i;

	/* Initialize encoding table */
	for(i = 0; i < NUMVALS; ++i) {
		self->codelen[i] = 0;
	}
}
/*  */
/* Recursive routine to walk the indicated subtree and level
 * and maintain the current path code in bstree. When a leaf
 * is found the entire code string and length are put into
 * the encoding table entry for the leaf's data value.
 *
 * Returns ERROR if codes are too long.
 */

int buildenc(SQ_COMPRESS_DATA *self,
int level, /* level of tree being examined, from zero */
int root) /* root of subtree is also data value if leaf */
{
	int l, r;

	l = self->node[root].lchild;
	r = self->node[root].rchild;

	if( l == NOCHILD && r == NOCHILD) {
		/* Leaf. Previous path determines bit string
		 * code of length level (bits 0 to level - 1).
		 * Ensures unused code bits are zero.
		 */
		self->codelen[root] = level;
		self->code[root] = self->tcode & ((unsigned int)(~0) >> ((sizeof(int) * 8) - level));
		return (level > 16) ? ERROR : 0;
	} else {
		if( l != NOCHILD) {
			/* Clear path bit and continue deeper */
			self->tcode &= ~(1 << level);
			/* NOTE RECURSION */
			if(buildenc(self, level + 1, l) == ERROR)
				return ERROR;
		}
		if(r != NOCHILD) {
			/* Set path bit and continue deeper */
			self->tcode |= 1 << level;
			/* NOTE RECURSION */
			if(buildenc(self, level + 1, r) == ERROR)
				return ERROR;
		}
	}
	return 0;	/* if we got here we're ok so far */
}
/*  */
/* Write out the header of the compressed file */

dsk_err_t wrt_head(SQ_COMPRESS_DATA *self, FILE *ob)
{
	int i, k, l, r;
	int numnodes;		/* nbr of nodes in simplified tree */
	dsk_err_t err;

	err = putwe(self, RECOGNIZE, ob);	/* identifies as compressed */
	if (err) return err;
	err = putwe(self, self->crc, ob);		/* unsigned sum of original data */
	if (err) return err;

	/* Record the original file name. */

	fwrite(self->sq_truename, 1, 1 + strlen(self->sq_truename), ob);	

	/* Write out a simplified decoding tree. Only the interior
	 * nodes are written. When a child is a leaf index
	 * (representing a data value) it is recoded as
	 * -(index + 1) to distinguish it from interior indexes
	 * which are recoded as positive indexes in the new tree.
	 * Note that this tree will be empty for an empty file.
	 */

	numnodes = self->dctreehd < NUMVALS ? 0 : self->dctreehd - (NUMVALS -1);
	err = putwe(self, numnodes, ob);
	if (err) return err;

	for(k = 0, i = self->dctreehd; k < numnodes; ++k, --i) {
		l = self->node[i].lchild;
		r = self->node[i].rchild;
		l = l < NUMVALS ? -(l + 1) : self->dctreehd - l;
		r = r < NUMVALS ? -(r + 1) : self->dctreehd - r;
		err = putwe(self, l, ob);	/* left child */
		if (err) return err;
		err = putwe(self, r, ob);	/* right child */
		if (err) return err;
	}
	return DSK_ERR_OK;
}
/*  */
/* Get an encoded byte or EOF. Reads from specified stream AS NEEDED.
 *
 * There are two unsynchronized bit-byte relationships here.
 * The input stream bytes are converted to bit strings of
 * various lengths via the static variables named c...
 * These bit strings are concatenated without padding to
 * become the stream of encoded result bytes, which this
 * function returns one at a time. The EOF (end of file) is
 * converted to SPEOF for convenience and encoded like any
 * other input value. True EOF is returned after that.
 *
 * The original gethuff() called a seperate function,
 * getbit(), but that more readable version was too slow.
 */

int gethuff(SQ_COMPRESS_DATA *self, FILE *fpin)	/*  Returns byte values except for EOF */
{
	unsigned int rbyte;	/* Result byte value */
	char need;		/* numbers of bits */

	rbyte = 0;
	need = 8;		/* build one byte per call */

	/* Loop to build a byte of encoded data
	 * Initialization forces read the first time
	 */

loop:
	if(self->cbitsrem >= need) {
		/* Current code fullfills our needs */
		if(need == 0)
			return rbyte & 0xff;
		/* Take what we need */
 		rbyte |= self->ccode << (8 - need);
		/* And leave the rest */
		self->ccode >>= need;
		self->cbitsrem -= need;
		return rbyte & 0xff;
	}

	/* We need more than current code */
	if(self->cbitsrem > 0) {
		/* Take what there is */
		rbyte |= self->ccode << (8 - need);
		need -= self->cbitsrem;
	}
	/* No more bits in current code string */
	if(self->curin == SPEOF) {
		/* The end of file token has been encoded. If
		 * result byte has data return it and do EOF next time
		 */
		self->cbitsrem = 0;

		return (need == 8) ? EOF : rbyte & 0xff;
	}

	/* Get an input byte */
	if((self->curin = getcnr(self, fpin)) == EOF)
		self->curin = SPEOF;	/* convenient for encoding */

	/* Get the new byte's code */
	self->ccode = self->code[self->curin];
	self->cbitsrem = self->codelen[self->curin];

	goto loop;
}


dsk_err_t putwe(SQ_COMPRESS_DATA *self, int w, FILE *iob)
{
	if (fputc(w & 0xFF, iob) == EOF) return DSK_ERR_SYSERR;
	if (fputc(w >> 8  , iob) == EOF) return DSK_ERR_SYSERR;
	return DSK_ERR_OK;
}


